
#define DIRECTINPUT_VERSION	0x0800
#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <dinput.h>
#include "common/GMLuaModule.h"

#pragma comment(lib, "dxguid.lib")
#pragma comment(lib, "dinput8.lib")

GMOD_MODULE( Init, Quit );

#define CHECK_ARG( POS, TYPE, STR ) if ( g_Lua->GetType( POS ) != TYPE ) { \
	char buf[ 255 ];\
	sprintf( buf, "%s expected, got %s\n", STR, g_Lua->GetType( POS ) == GLua::TYPE_NIL ? "no value" : g_Lua->GetTypeName( POS ) );\
	g_Lua->LuaError( buf, POS ); return 0;}


HWND g_MainWindow = 0;
LPDIRECTINPUT g_lpDirectInput = 0;
LPDIRECTINPUTDEVICE g_lpKeyboard = 0;
unsigned char g_KeyState[256] = {0};

void DxQuit()
{
	if (g_lpKeyboard)
	{
		g_lpKeyboard->Release();
		g_lpKeyboard = 0;
	}
	if (g_lpDirectInput)
	{
		g_lpDirectInput->Release();
		g_lpDirectInput = 0;
	}
}

BOOL CALLBACK FindTheWindow( HWND hwnd, LPARAM lParam )
{
	HMODULE hToFind = (HMODULE)lParam;
	HMODULE hModule = (HMODULE)(LONG_PTR)GetWindowLongPtr(hwnd, GWL_HINSTANCE);

	if (hToFind == hModule)
	{
		g_MainWindow = hwnd;
		return FALSE;		// Found the main window, stop enumeration!
	}
	
	return TRUE;
}

bool FindMainWindow()
{
	HMODULE hModule = GetModuleHandle(NULL);

	EnumWindows(FindTheWindow, (LPARAM)hModule);

	if (g_MainWindow)
		return true;

	return false;
}

bool DxInit()
{
	if (!FindMainWindow())
	{
		Msg( "Couldn't find the main window! :(\n" );
		return false;
	}

	HRESULT hr;

	if (FAILED(hr = DirectInput8Create(GetModuleHandle(NULL), DIRECTINPUT_VERSION, IID_IDirectInput8, (void**)&g_lpDirectInput, NULL)))
	{
		Msg( "DirectInput8Create failed! :(\n" );
		//Msg( "DirectX said: %s\n", DXGetErrorString8(hr) );
		return false;
	}

	if (FAILED(hr = g_lpDirectInput->CreateDevice(GUID_SysKeyboard, &g_lpKeyboard, NULL)))
	{
		Msg( "g_lpDirectInput->CreateDevice failed! :(\n" );
		//Msg( "DirectX said: %s\n", DXGetErrorString8(hr) );
		g_lpDirectInput->Release();
		g_lpDirectInput = 0;
		return false;
	}
  
	if (FAILED(hr = g_lpKeyboard->SetDataFormat(&c_dfDIKeyboard)))
	{ 
		Msg( "g_lpKeyboard->SetDataFormat failed! :(\n" );
		//Msg( "DirectX said: %s\n", DXGetErrorString8(hr) );
		g_lpKeyboard->Release();
		g_lpKeyboard = 0;
		g_lpDirectInput->Release();
		g_lpDirectInput = 0;
		return false;
	}
  
	if (FAILED(hr = g_lpKeyboard->SetCooperativeLevel(NULL, DISCL_BACKGROUND | DISCL_NONEXCLUSIVE)))
	{
		Msg( "g_lpKeyboard->SetCooperativeLevel failed! :(\n" );
		//Msg( "DirectX said: %s\n", DXGetErrorString8(hr) );
		g_lpKeyboard->Release();
		g_lpKeyboard = 0;
		g_lpDirectInput->Release();
		g_lpDirectInput = 0;
		return false;
	}
  
	if (FAILED(hr = g_lpKeyboard->Acquire()))
	{
		Msg( "g_lpKeyboard->Acquire failed! :(\n" );
		//Msg( "DirectX said: %s\n", DXGetErrorString8(hr) );
		g_lpKeyboard->Release();
		g_lpKeyboard = 0;
		g_lpDirectInput->Release();
		g_lpDirectInput = 0;
		return false;
	}

	Msg( "DxInit succeeded! huzzah! :)\n" );
	return true;
}

LUA_FUNCTION( LIB_GetKeyState )
{
	if (g_lpKeyboard == 0)
	{
		return 0;
	}

	if (FAILED(g_lpKeyboard->GetDeviceState(sizeof(g_KeyState), (LPVOID)g_KeyState)))
	{
		Msg( "g_lpKeyboard->GetDeviceState failed! :(\n" );
		return 0;
	}

	return 0;
}

LUA_FUNCTION( LIB_AllKeyState )
{
	ILuaObject* pKeyState = g_Lua->GetNewTable();
	
	for (int i=0; i<256; i++)
		pKeyState->SetMember( (float)i, (g_KeyState[i] & 0x80) != 0 );

	g_Lua->Push( pKeyState );
	SAFE_UNREF(pKeyState)

	return 1;
}

LUA_FUNCTION( LIB_KeyUp )
{
	CHECK_ARG(1, GLua::TYPE_NUMBER, "Number");

	int nKey = g_Lua->GetInteger(1);

	if (nKey < 0 || nKey >= 256)
	{
		g_Lua->Push(false);
		return 1;
	}

	g_Lua->Push( (g_KeyState[nKey] & 0x80) == 0 );
	return 1;
}

LUA_FUNCTION( LIB_KeyDown )
{
	CHECK_ARG(1, GLua::TYPE_NUMBER, "Number");

	int nKey = g_Lua->GetInteger(1);

	if (nKey < 0 || nKey >= 256)
	{
		g_Lua->Push(false);
		return 1;
	}

	g_Lua->Push( (g_KeyState[nKey] & 0x80) != 0 );
	return 1;
}

#define KEY_BIND( dik )	pKeyboard->SetMember(#dik, (float)dik);

int Init( void )
{
	ILuaObject* pKeyboard = g_Lua->GetNewTable();
		
	pKeyboard->SetMember( "InitOk",			DxInit() );

	pKeyboard->SetMember( "GetKeyState",	LIB_GetKeyState );
	pKeyboard->SetMember( "KeyDown",		LIB_KeyDown );
	pKeyboard->SetMember( "KeyUp",			LIB_KeyUp );
	pKeyboard->SetMember( "AllKeys",		LIB_AllKeyState );

		KEY_BIND(    DIK_0            )
		KEY_BIND(    DIK_1            )
		KEY_BIND(    DIK_2            )
		KEY_BIND(    DIK_3            )
		KEY_BIND(    DIK_4            )
		KEY_BIND(    DIK_5            )
		KEY_BIND(    DIK_6            )
		KEY_BIND(    DIK_7            )
		KEY_BIND(    DIK_8            )
		KEY_BIND(    DIK_9            )
		KEY_BIND(    DIK_A            )
		KEY_BIND(    DIK_ABNT_C1      )
		KEY_BIND(    DIK_ABNT_C2      )
		KEY_BIND(    DIK_ADD          )
		KEY_BIND(    DIK_APOSTROPHE   )
		KEY_BIND(    DIK_APPS         )
		KEY_BIND(    DIK_AT           )
		KEY_BIND(    DIK_AX           )
		KEY_BIND(    DIK_B            )
		KEY_BIND(    DIK_BACK         )
		KEY_BIND(    DIK_BACKSLASH    )
		KEY_BIND(    DIK_BACKSPACE    )
		KEY_BIND(    DIK_C            )
		KEY_BIND(    DIK_CALCULATOR   )
		KEY_BIND(    DIK_CAPITAL      )
		KEY_BIND(    DIK_CAPSLOCK     )
		KEY_BIND(    DIK_CIRCUMFLEX   )
		KEY_BIND(    DIK_COLON        )
		KEY_BIND(    DIK_COMMA        )
		KEY_BIND(    DIK_CONVERT      )
		KEY_BIND(    DIK_D            )
		KEY_BIND(    DIK_DECIMAL      )
		KEY_BIND(    DIK_DELETE       )
		KEY_BIND(    DIK_DIVIDE       )
		KEY_BIND(    DIK_DOWN         )
		KEY_BIND(    DIK_DOWNARROW    )
		KEY_BIND(    DIK_E            )
		KEY_BIND(    DIK_END          )
		KEY_BIND(    DIK_EQUALS       )
		KEY_BIND(    DIK_ESCAPE       )
		KEY_BIND(    DIK_F            )
		KEY_BIND(    DIK_F1           )
		KEY_BIND(    DIK_F2           )
		KEY_BIND(    DIK_F3           )
		KEY_BIND(    DIK_F4           )
		KEY_BIND(    DIK_F5           )
		KEY_BIND(    DIK_F6           )
		KEY_BIND(    DIK_F7           )
		KEY_BIND(    DIK_F8           )
		KEY_BIND(    DIK_F9           )
		KEY_BIND(    DIK_F10          )
		KEY_BIND(    DIK_F11          )
		KEY_BIND(    DIK_F12          )
		KEY_BIND(    DIK_F13          )
		KEY_BIND(    DIK_F14          )
		KEY_BIND(    DIK_F15          )
		KEY_BIND(    DIK_G            )
		KEY_BIND(    DIK_GRAVE        )
		KEY_BIND(    DIK_H            )
		KEY_BIND(    DIK_HOME         )
		KEY_BIND(    DIK_I            )
		KEY_BIND(    DIK_INSERT       )
		KEY_BIND(    DIK_J            )
		KEY_BIND(    DIK_K            )
		KEY_BIND(    DIK_KANA         )
		KEY_BIND(    DIK_KANJI        )
		KEY_BIND(    DIK_L            )
		KEY_BIND(    DIK_LALT         )
		KEY_BIND(    DIK_LBRACKET     )
		KEY_BIND(    DIK_LCONTROL     )
		KEY_BIND(    DIK_LEFT         )
		KEY_BIND(    DIK_LEFTARROW    )
		KEY_BIND(    DIK_LMENU        )
		KEY_BIND(    DIK_LSHIFT       )
		KEY_BIND(    DIK_LWIN         )
		KEY_BIND(    DIK_M            )
		KEY_BIND(    DIK_MAIL         )
		KEY_BIND(    DIK_MEDIASELECT  )
		KEY_BIND(    DIK_MEDIASTOP    )
		KEY_BIND(    DIK_MINUS        )
		KEY_BIND(    DIK_MULTIPLY     )
		KEY_BIND(    DIK_MUTE         )
		KEY_BIND(    DIK_MYCOMPUTER   )
		KEY_BIND(    DIK_N            )
		KEY_BIND(    DIK_NEXT         )
		KEY_BIND(    DIK_NEXTTRACK    )
		KEY_BIND(    DIK_NOCONVERT    )
		KEY_BIND(    DIK_NUMLOCK      )
		KEY_BIND(    DIK_NUMPAD0      )
		KEY_BIND(    DIK_NUMPAD1      )
		KEY_BIND(    DIK_NUMPAD2      )
		KEY_BIND(    DIK_NUMPAD3      )
		KEY_BIND(    DIK_NUMPAD4      )
		KEY_BIND(    DIK_NUMPAD5      )
		KEY_BIND(    DIK_NUMPAD6      )
		KEY_BIND(    DIK_NUMPAD7      )
		KEY_BIND(    DIK_NUMPAD8      )
		KEY_BIND(    DIK_NUMPAD9      )
		KEY_BIND(    DIK_NUMPADCOMMA  )
		KEY_BIND(    DIK_NUMPADENTER  )
		KEY_BIND(    DIK_NUMPADEQUALS )
		KEY_BIND(    DIK_NUMPADMINUS  )
		KEY_BIND(    DIK_NUMPADPERIOD )
		KEY_BIND(    DIK_NUMPADPLUS   )
		KEY_BIND(    DIK_NUMPADSLASH  )
		KEY_BIND(    DIK_NUMPADSTAR   )
		KEY_BIND(    DIK_O            )
		KEY_BIND(    DIK_OEM_102      )
		KEY_BIND(    DIK_P            )
		KEY_BIND(    DIK_PAUSE        )
		KEY_BIND(    DIK_PERIOD       )
		KEY_BIND(    DIK_PGDN         )
		KEY_BIND(    DIK_PGUP         )
		KEY_BIND(    DIK_PLAYPAUSE    )
		KEY_BIND(    DIK_POWER        )
		KEY_BIND(    DIK_PREVTRACK    )
		KEY_BIND(    DIK_PRIOR        )
		KEY_BIND(    DIK_Q            )
		KEY_BIND(    DIK_R            )
		KEY_BIND(    DIK_RALT         )
		KEY_BIND(    DIK_RBRACKET     )
		KEY_BIND(    DIK_RCONTROL     )
		KEY_BIND(    DIK_RETURN       )
		KEY_BIND(    DIK_RIGHT        )
		KEY_BIND(    DIK_RIGHTARROW   )
		KEY_BIND(    DIK_RMENU        )
		KEY_BIND(    DIK_RSHIFT       )
		KEY_BIND(    DIK_RWIN         )
		KEY_BIND(    DIK_S            )
		KEY_BIND(    DIK_SCROLL       )
		KEY_BIND(    DIK_SEMICOLON    )
		KEY_BIND(    DIK_SLASH        )
		KEY_BIND(    DIK_SLEEP        )
		KEY_BIND(    DIK_SPACE        )
		KEY_BIND(    DIK_STOP         )
		KEY_BIND(    DIK_SUBTRACT     )
		KEY_BIND(    DIK_SYSRQ        )
		KEY_BIND(    DIK_T            )
		KEY_BIND(    DIK_TAB          )
		KEY_BIND(    DIK_U            )
		KEY_BIND(    DIK_UNDERLINE    )
		KEY_BIND(    DIK_UNLABELED    )
		KEY_BIND(    DIK_UP           )
		KEY_BIND(    DIK_UPARROW      )
		KEY_BIND(    DIK_V            )
		KEY_BIND(    DIK_VOLUMEDOWN   )
		KEY_BIND(    DIK_VOLUMEUP     )
		KEY_BIND(    DIK_W            )
		KEY_BIND(    DIK_WAKE         )
		KEY_BIND(    DIK_WEBBACK      )
		KEY_BIND(    DIK_WEBFAVORITES )
		KEY_BIND(    DIK_WEBFORWARD   )
		KEY_BIND(    DIK_WEBHOME      )
		KEY_BIND(    DIK_WEBREFRESH   )
		KEY_BIND(    DIK_WEBSEARCH    )
		KEY_BIND(    DIK_WEBSTOP      )
		KEY_BIND(    DIK_X            )
		KEY_BIND(    DIK_Y            )
		KEY_BIND(    DIK_YEN          )
		KEY_BIND(    DIK_Z            )

	g_Lua->SetGlobal("keyboard", pKeyboard);

	SAFE_UNREF(pKeyboard);
	return 1;
}

int Quit( void )
{
	DxQuit();
	return 1;
}


